/* Copyright (c) 2008 PrimeBase Technologies GmbH, Germany
 *
 * PrimeBase Media Stream for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Original author: Paul McCullagh
 * Continued development: Barry Leslie
 *
 * 2007-06-04
 *
 * H&G2JCtL
 *
 * Media Stream Tables.
 *
 */
#include "CSConfig.h"

#include "CSGlobal.h"
#include "CSLog.h"
#include "CSStrUtil.h"
#include "CSPath.h"

#include "OpenTable_ms.h"
#include "Table_ms.h"
#include "ConnectionHandler_ms.h"
#include "Engine_ms.h"
#include "Util_ms.h"
#include "Transaction_ms.h"

#include "ms_mysql.h"

/*
 * ---------------------------------------------------------------
 * OPEN TABLES
 */

MSOpenTable::MSOpenTable():
CSRefObject(),
CSPooled(),
inUse(true),
isNotATable(false),
nextTable(NULL),
myPool(NULL),
//myBlobType(MS_CLOUD_PENDING_STORAGE),
myBlobType(MS_STANDARD_STORAGE),
myTableFile(NULL),
myWriteRepo(NULL),
myWriteRepoFile(NULL),
myTempLogFile(NULL),
iNextLink(NULL),
iPrevLink(NULL)
//iUseSize(0),
//iUseCount(0),
//iUsedBlobs(0)
{
}

MSOpenTable::~MSOpenTable()
{
	close();
}

void MSOpenTable::close()
{
	enter_();
	if (myTableFile) {
		myTableFile->release();
		myTableFile = NULL;
	}
	closeForWriting();
	if (myTempLogFile) {
		myTempLogFile->release();
		myTempLogFile = NULL;
	}
/*
	if (iUsedBlobs) {
		cs_free(iUsedBlobs);
		iUsedBlobs = NULL;
	}
	iUseCount = 0;
	iUseSize = 0;
*/
	exit_();
}

void MSOpenTable::returnToPool()
{
	MSTableList::releaseTable(this);
}

void MSOpenTable::createBlob(PBMSBlobURLPtr bh, csWord8 blob_size, char *metadata, csWord2 metadata_size, CSInputStream *stream)
{
	csWord8 repo_offset;
	csWord8 blob_id = 0;
	csWord4	auth_code;
	csWord2 head_size;
	csWord4	log_id;
	csWord4 log_offset;
	csWord4 temp_time;
	csWord8 repo_size;
	csWord8 repo_id;
	Md5Digest checksum;
	enter_();
	
	push_(stream);
	openForWriting();
	ASSERT(myWriteRepo);
	auth_code = random();
	repo_size = myWriteRepo->getRepoFileSize();
	temp_time =	myWriteRepo->myLastTempTime;

	try_(a) {
		pop_(stream);
		repo_offset = myWriteRepo->receiveBlob(this, &head_size, blob_size, metadata_size, &checksum, stream);
		
		repo_id = myWriteRepo->myRepoID;
		if (isNotATable) {	
			getDB()->queueForDeletion(this, MS_TL_REPO_REF, repo_id, repo_offset, auth_code, &log_id, &log_offset, &temp_time, NULL);
			formatRepoURL(bh->bu_data, repo_id, repo_offset, auth_code, blob_size);
		}
		else {
			blob_id = getDBTable()->createBlobHandle(this, myWriteRepo->myRepoID, repo_offset, blob_size, head_size, auth_code);
			getDB()->queueForDeletion(this, MS_TL_BLOB_REF, getDBTable()->myTableID, blob_id, auth_code, &log_id, &log_offset, &temp_time, NULL);
			formatBlobURL(bh->bu_data, blob_id, auth_code, blob_size, 0);
		}
		
		myWriteRepo->writeBlobHead(this, repo_offset, myWriteRepo->myRepoDefRefSize, head_size, blob_size, &checksum, metadata, metadata_size, blob_id, auth_code, log_id, log_offset, myBlobType);
	}
	
	catch_(a);
	// In the event of an error reset the repository size to its original size.
	myWriteRepo->setRepoFileSize(this, repo_size);
	throw_();
	
	cont_(a);
	exit_();
}

// BLOBs created with this method are always created as standard local BLOBs. (No cloud storage)
void MSOpenTable::createBlob(PBMSBlobIDPtr blob_id, csWord8 blob_size, char *metadata, csWord2 metadata_size)
{
	csWord8 repo_size;
	csWord8 repo_offset;
	csWord8 repo_id;
	csWord4	auth_code;
	csWord2 head_size;
	csWord4	log_id;
	csWord4 log_offset;
	csWord4 temp_time;
	enter_();

	openForWriting();
	ASSERT(myWriteRepo);
	auth_code = random();
	
	repo_size = myWriteRepo->getRepoFileSize();
	try_(a) {
		repo_offset = myWriteRepo->receiveBlob(this, &head_size, blob_size, metadata_size);
		repo_id = myWriteRepo->myRepoID;
		temp_time = myWriteRepo->myLastTempTime;
		getDB()->queueForDeletion(this, MS_TL_REPO_REF, repo_id, repo_offset, auth_code, &log_id, &log_offset, &temp_time, NULL);
		myWriteRepo->myLastTempTime = temp_time;
		myWriteRepo->writeBlobHead(this, repo_offset, myWriteRepo->myRepoDefRefSize, head_size, blob_size, NULL, metadata, metadata_size, 0, auth_code, log_id, log_offset, MS_STANDARD_STORAGE);
		// myWriteRepo->setRepoFileSize(this, repo_offset + head_size + blob_size);This is now set by writeBlobHead()
		
		blob_id->bi_db_id = getDB()->myDatabaseID;
		blob_id->bi_blob_id = repo_offset;
		blob_id->bi_tab_id = repo_id;
		blob_id->bi_auth_code = auth_code;
		blob_id->bi_blob_size = blob_size;
		blob_id->bi_blob_type = MS_URL_TYPE_REPO;
		blob_id->bi_blob_ref_id = 0;
	
	}
	
	catch_(a);
	// BLOBs created with this method are always created as standard local BLOBs. (No cloud storage)
	myWriteRepo->setRepoFileSize(this, repo_size);
	throw_();
	
	cont_(a);
	exit_();
}

void MSOpenTable::sendRepoBlob(csWord8 blob_id, csWord4 auth_code, bool info_only, CSHTTPOutputStream *stream)
{
	csWord4		repo_id;
	csWord8		offset;
	csWord8		size;
	csWord2		head_size;
	MSRepoFile	*repo_file;

	enter_();
	openForReading();
	getDBTable()->readBlobHandle(this, blob_id, &auth_code, &repo_id, &offset, &size, &head_size, true);
	repo_file = getDB()->getRepoFileFromPool(repo_id, false);
	frompool_(repo_file);
	//repo_file->sendBlob(this, offset, head_size, size, stream);
	repo_file->sendBlob(this, offset, 0, false, info_only, stream);
	backtopool_(repo_file);
	exit_();
}

/* OLD WAY
void MSOpenTable::useBlob(int type, csWord4 db_id, csWord4 tab_id, csWord8 blob_id, csWord4 auth_code, csWord2 col_index, csWord8 blob_size, char **ret_blob_url)
{
	MSUsedBlobPtr	ub;
	MSRepoFile		*repo_file;
	MSBlobHeadRec	blob;
	CSInputStream	*stream;
	MSDatabase		*blob_db;

	enter_();
	if (iUseCount == iUseSize) {
		cs_realloc((void **) &iUsedBlobs, sizeof(MSUsedBlobRec) * (iUseSize+1));
		iUseSize++;
	}
	ub = &iUsedBlobs[iUseCount];
	
	blob_db = getDB();
		
	if (blob_db->isRecovering()) {
		// During recovery the only thing that needs to be done is to 
		// reset the database ID which is done when the URL is created.
		// Create the URL using the table ID passed in not the one from 
		// the table associated with this object.
		ub->ub_blob_id = blob_id;
		ub->ub_blob_size = blob_size;
		formatBlobURL(ub->ub_blob_url, ub->ub_blob_id, auth_code, ub->ub_blob_size, tab_id, ub->ub_blob_ref_id);
	} else {
		openForReading();
		if (type == MS_URL_TYPE_REPO) { // There is no table reference associated with this BLOB yet.
			csWord4		ac;
			csWord1		status;
			bool		same_db = true;

			if (blob_db->myDatabaseID == db_id)
				repo_file = blob_db->getRepoFileFromPool(tab_id, false);
			else {
				same_db = false;
				blob_db = MSDatabase::getDatabase(db_id);
				push_(blob_db);
				repo_file = blob_db->getRepoFileFromPool(tab_id, false);
				release_(blob_db);
				blob_db = repo_file->myRepo->myRepoDatabase;
			}
		
			frompool_(repo_file);
			repo_file->read(&blob, blob_id, MS_MIN_BLOB_HEAD_SIZE, MS_MIN_BLOB_HEAD_SIZE);

			ub->ub_repo_offset = blob_id;
			ub->ub_blob_size  = CS_GET_DISK_6(blob.rb_blob_size_6);
			ub->ub_head_size = CS_GET_DISK_2(blob.rb_head_size_2);
			 
			ac = CS_GET_DISK_4(blob.rb_auth_code_4);
			if (auth_code != ac)
				CSException::throwException(CS_CONTEXT, MS_ERR_AUTH_FAILED, "Invalid BLOB identifier");
			status = CS_GET_DISK_1(blob.rb_status_1);
			if ( ! IN_USE_BLOB_STATUS(status))
				CSException::throwException(CS_CONTEXT, MS_ERR_NOT_FOUND, "BLOB has already been deleted");

			if (same_db) {
				// Create a table reference to the BLOB:
				ub->ub_repo_id = tab_id;
				ub->ub_blob_id = getDBTable()->createBlobHandle(this, tab_id, blob_id, ub->ub_blob_size, ub->ub_head_size, auth_code);
				ub->ub_state = MS_UB_NEW_HANDLE;
			}
			else {
				
				getDB()->openWriteRepo(this);

				// If either databases are using cloud storage then this is
				// not supported yet.			
				if (getDB()->myBlobCloud || myWriteRepo->myRepoDatabase->myBlobCloud)
					CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Copying cloud BLOB between databases is not supported.");
			
				stream = repo_file->getInputStream(ub->ub_repo_offset);
				push_(stream);
				ub->ub_repo_offset = myWriteRepo->copyBlob(this, ub->ub_head_size + ub->ub_blob_size, stream);			
				release_(stream);

				// Create a table reference to the BLOB:
				ub->ub_repo_id = myWriteRepo->myRepoID;
				ub->ub_blob_id = getDBTable()->createBlobHandle(this, myWriteRepo->myRepoID, ub->ub_repo_offset, ub->ub_blob_size, ub->ub_head_size, auth_code);
				ub->ub_state = MS_UB_NEW_BLOB;
			}
			backtopool_(repo_file);
		}
		else {

			if (blob_db->myDatabaseID == db_id && getDBTable()->myTableID == tab_id) {
				getDBTable()->readBlobHandle(this, blob_id, &auth_code, &ub->ub_repo_id, &ub->ub_repo_offset, &ub->ub_blob_size, &ub->ub_head_size, true);
				
				ub->ub_blob_id = blob_id;
				ub->ub_state = MS_UB_SAME_TAB;
			}
			else {
				MSOpenTable *blob_otab;

				blob_otab = MSTableList::getOpenTableByID(db_id, tab_id);
				frompool_(blob_otab);
				blob_otab->getDBTable()->readBlobHandle(blob_otab, blob_id, &auth_code, &ub->ub_repo_id, &ub->ub_repo_offset, &ub->ub_blob_size, &ub->ub_head_size, true);
				if (blob_db->myDatabaseID == db_id) {
					ub->ub_blob_id = getDBTable()->createBlobHandle(this, ub->ub_repo_id, ub->ub_repo_offset, ub->ub_blob_size, ub->ub_head_size, auth_code);
					ub->ub_state = MS_UB_NEW_HANDLE;
				}
				else {

					blob_db->openWriteRepo(this);
									
					stream = repo_file->getInputStream(ub->ub_repo_offset);
					push_(stream);
					
					ub->ub_repo_offset = myWriteRepo->copyBlob(this, ub->ub_head_size + ub->ub_blob_size, stream);
					
					release_(stream);

					ub->ub_repo_id = myWriteRepo->myRepoID;
					ub->ub_blob_id = getDBTable()->createBlobHandle(this, myWriteRepo->myRepoID, ub->ub_repo_offset, ub->ub_blob_size, ub->ub_head_size, auth_code);
					ub->ub_state = MS_UB_NEW_BLOB;
				}
				backtopool_(blob_otab);
			}
		}
		
		ub->ub_blob_ref_id = blob_db->newBlobRefId();
		formatBlobURL(ub->ub_blob_url, ub->ub_blob_id, auth_code, ub->ub_blob_size, ub->ub_blob_ref_id);
	} 
		
	ub->ub_auth_code = auth_code;
	ub->ub_col_index = col_index;
	*ret_blob_url = ub->ub_blob_url;
	iUseCount++;
	exit_();
}

void MSOpenTable::unUseBlobs()
{
	MSUsedBlobPtr	ub = iUsedBlobs;

	if (!getDB()->isRecovering()) {
		for (u_int i=0; i<iUseCount; i++) {
			switch (ub->ub_state) {
				case MS_UB_SAME_TAB:
					break;
				case MS_UB_NEW_HANDLE:
				case MS_UB_NEW_BLOB:
					getDBTable()->freeBlobHandle(this, ub->ub_blob_id, ub->ub_repo_id, ub->ub_repo_offset, ub->ub_auth_code);
					break;
				case MS_UB_RETAINED:
				case MS_UB_NEW_RETAINED:
					freeReference(ub->ub_blob_id, ub->ub_blob_ref_id);
					break;
			}
			ub++;
		}
	}
	iUseCount = 0;
}

void MSOpenTable::retainReferences()
{
	MSUsedBlobPtr	ub = iUsedBlobs;
	off_t			repo_offset = 0;
	csWord4			tab_id;

	enter_();
	if (!getDB()->isRecovering()) {
	
		tab_id = getDBTable()->myTableID;
		
		for (u_int i=0; i<iUseCount; i++) {
			if (ub->ub_state == MS_UB_NEW_BLOB) {
				// In this case a blob has been copied because it was referenced from a database
				// other than the one in which it resides. The BLOB and header has been written to the repository 
				// of the destination database but the header has not been updated yet.
				myWriteRepo->resetBlobHead(this, ub->ub_repo_offset, ub->ub_head_size, ub->ub_blob_id, ub->ub_blob_ref_id, ub->ub_auth_code, ub->ub_col_index, myBlobType);
				if (repo_offset < ub->ub_repo_offset + ub->ub_head_size +  ub->ub_blob_size)
					repo_offset = ub->ub_repo_offset + ub->ub_head_size +  ub->ub_blob_size;
				ub->ub_state = MS_UB_NEW_RETAINED;
			}
			else {
				MSRepoFile *repo_file;

				repo_file = getDB()->getRepoFileFromPool(ub->ub_repo_id, false);
				frompool_(repo_file);
				repo_file->retainBlob(this, ub->ub_repo_offset, ub->ub_head_size, tab_id, ub->ub_blob_id, ub->ub_blob_ref_id, ub->ub_auth_code, ub->ub_col_index);
				backtopool_(repo_file);
				ub->ub_state = MS_UB_RETAINED;
			}
			
			MSTransactionManager::referenceBLOB(getDB()->myDatabaseID, tab_id, ub->ub_blob_id, ub->ub_blob_ref_id);
			ub++;
		}
	}

	iUseCount = 0;
	if (repo_offset)
		myWriteRepo->setRepoFileSize(this, repo_offset);
	exit_();
}
*/

void MSOpenTable::freeReference(csWord8 blob_id, csWord8 blob_ref_id)
{
	csWord4		repo_id;
	csWord8		offset;
	csWord8		blob_size;
	csWord2		head_size;
	MSRepoFile	*repo_file;
	csWord4		auth_code = 0;

	enter_();
	openForReading();

	getDBTable()->readBlobHandle(this, blob_id, &auth_code, &repo_id, &offset, &blob_size, &head_size, true);
	repo_file = getDB()->getRepoFileFromPool(repo_id, false);

	frompool_(repo_file);
	repo_file->releaseBlob(this, offset, head_size, getDBTable()->myTableID, blob_id, blob_ref_id, auth_code);
	backtopool_(repo_file);

	exit_();
}

void MSOpenTable::commitReference(csWord8 blob_id, csWord8 blob_ref_id)
{
	csWord4		repo_id;
	csWord8		offset;
	csWord8		blob_size;
	csWord2		head_size;
	MSRepoFile	*repo_file;
	csWord4		auth_code = 0;

	enter_();
	openForReading();
	
	getDBTable()->readBlobHandle(this, blob_id, &auth_code, &repo_id, &offset, &blob_size, &head_size, true);
	repo_file = getDB()->getRepoFileFromPool(repo_id, false);

	frompool_(repo_file);
	repo_file->commitBlob(this, offset, head_size, getDBTable()->myTableID, blob_id, blob_ref_id, auth_code);
	backtopool_(repo_file);

	exit_();
}

void MSOpenTable::useBlob(int type, csWord4 db_id, csWord4 tab_id, csWord8 blob_id, csWord4 auth_code, csWord2 col_index, csWord8 blob_size, csWord8 blob_ref_id, char *ret_blob_url)
{
	MSRepoFile		*repo_file;
	MSBlobHeadRec	blob;
	CSInputStream	*stream;
	MSDatabase		*blob_db;
	int				state;
	csWord2			head_size;
	csWord8			repo_offset;
	csWord4			repo_id;

	enter_();

	blob_db = getDB();
		
	if (!blob_db->isRecovering()) {
		// During recovery the only thing that needs to be done is to 
		// reset the database ID which is done when the URL is created.
		// Create the URL using the table ID passed in not the one from 
		// the table associated with this object.

		openForReading();
		if (type == MS_URL_TYPE_REPO) { // There is no table reference associated with this BLOB yet.
			csWord4		ac;
			csWord1		status;
			bool		same_db = true;

			if (blob_db->myDatabaseID == db_id)
				repo_file = blob_db->getRepoFileFromPool(tab_id, false);
			else {
				same_db = false;
				blob_db = MSDatabase::getDatabase(db_id);
				push_(blob_db);
				repo_file = blob_db->getRepoFileFromPool(tab_id, false);
				release_(blob_db);
				blob_db = repo_file->myRepo->myRepoDatabase;
			}
		
			frompool_(repo_file);
			repo_file->read(&blob, blob_id, MS_MIN_BLOB_HEAD_SIZE, MS_MIN_BLOB_HEAD_SIZE);

			repo_offset = blob_id;
			blob_size  = CS_GET_DISK_6(blob.rb_blob_size_6);
			head_size = CS_GET_DISK_2(blob.rb_head_size_2);
			 
			ac = CS_GET_DISK_4(blob.rb_auth_code_4);
			if (auth_code != ac)
				CSException::throwException(CS_CONTEXT, MS_ERR_AUTH_FAILED, "Invalid BLOB identifier");
			status = CS_GET_DISK_1(blob.rb_status_1);
			if ( ! IN_USE_BLOB_STATUS(status))
				CSException::throwException(CS_CONTEXT, MS_ERR_NOT_FOUND, "BLOB has already been deleted");

			if (same_db) {
				// Create a table reference to the BLOB:
				repo_id = tab_id;
				blob_id = getDBTable()->createBlobHandle(this, tab_id, blob_id, blob_size, head_size, auth_code);
				state = MS_UB_NEW_HANDLE;
			}
			else {
				
				getDB()->openWriteRepo(this);

				// If either databases are using cloud storage then this is
				// not supported yet.			
				if (getDB()->myBlobCloud || myWriteRepo->myRepoDatabase->myBlobCloud)
					CSException::throwException(CS_CONTEXT, CS_ERR_GENERIC_ERROR, "Copying cloud BLOB between databases is not supported.");
			
				stream = repo_file->getInputStream(repo_offset);
				push_(stream);
				repo_offset = myWriteRepo->copyBlob(this, head_size + blob_size, stream);			
				release_(stream);

				// Create a table reference to the BLOB:
				repo_id = myWriteRepo->myRepoID;
				blob_id = getDBTable()->createBlobHandle(this, myWriteRepo->myRepoID, repo_offset, blob_size, head_size, auth_code);
				state = MS_UB_NEW_BLOB;
			}
			backtopool_(repo_file);
		}
		else {

			if (blob_db->myDatabaseID == db_id && getDBTable()->myTableID == tab_id) {
				getDBTable()->readBlobHandle(this, blob_id, &auth_code, &repo_id, &repo_offset, &blob_size, &head_size, true);
				
				state = MS_UB_SAME_TAB;
			}
			else {
				MSOpenTable *blob_otab;

				blob_otab = MSTableList::getOpenTableByID(db_id, tab_id);
				frompool_(blob_otab);
				blob_otab->getDBTable()->readBlobHandle(blob_otab, blob_id, &auth_code, &repo_id, &repo_offset, &blob_size, &head_size, true);
				if (blob_db->myDatabaseID == db_id) {
					blob_id = getDBTable()->findBlobHandle(this, repo_id, repo_offset, blob_size, head_size, auth_code);
					if (blob_id == 0)
						blob_id = getDBTable()->createBlobHandle(this, repo_id, repo_offset, blob_size, head_size, auth_code);
					state = MS_UB_NEW_HANDLE;
				}
				else {

					// NOTE: For each BLOB reference copied from one database to another a new
					// BLOB will be created. This can result in multiple copies fo the same BLOB
					// in the destination database. One way around this would be to redisign things
					// so that there is one BLOB repository shared across all databases. 
					blob_db->openWriteRepo(this);
									
					stream = repo_file->getInputStream(repo_offset);
					push_(stream);
					
					repo_offset = myWriteRepo->copyBlob(this, head_size + blob_size, stream);
					
					release_(stream);

					repo_id = myWriteRepo->myRepoID;
					blob_id = getDBTable()->createBlobHandle(this, myWriteRepo->myRepoID, repo_offset, blob_size, head_size, auth_code);
					state = MS_UB_NEW_BLOB;
				}
				backtopool_(blob_otab);
			}
			
		}
		
		blob_ref_id = blob_db->newBlobRefId();
		
		// Always use the table ID of this table because regardless of
		// where the BLOB ref came from it is being inserted into this table.
		tab_id = getDBTable()->myTableID; 
		
		// Add the BLOB reference to the repository.
		repo_file = blob_db->getRepoFileFromPool(repo_id, false);		
		frompool_(repo_file);
		repo_file->referenceBlob(this, repo_offset, head_size, tab_id, blob_id, blob_ref_id, auth_code, col_index);		
		backtopool_(repo_file);
		
		MSTransactionManager::referenceBLOB(getDB()->myDatabaseID, tab_id, blob_id, blob_ref_id);

	} 
	
	formatBlobURL(ret_blob_url, blob_id, auth_code, blob_size, tab_id, blob_ref_id);
		
	exit_();
}

void MSOpenTable::releaseReference(csWord8 blob_id, csWord8 blob_ref_id)
{
	enter_();
	
	MSTransactionManager::dereferenceBLOB(getDB()->myDatabaseID, getDBTable()->myTableID, blob_id, blob_ref_id);

	exit_();
}

void MSOpenTable::checkBlob(CSStringBuffer *buffer, csWord8 blob_id, csWord4 auth_code, csWord4 temp_log_id, csWord4 temp_log_offset)
{
	csWord4		repo_id;
	csWord8		offset;
	csWord8		size;
	csWord2		head_size;
	MSRepoFile	*repo_file;

	enter_();
	openForReading();
	if (getDBTable()->readBlobHandle(this, blob_id, &auth_code, &repo_id, &offset, &size, &head_size, false)) {
		if ((repo_file = getDB()->getRepoFileFromPool(repo_id, true))) {
			frompool_(repo_file);
			repo_file->checkBlob(this, buffer, offset, auth_code, temp_log_id, temp_log_offset);
			backtopool_(repo_file);
		}
		else
			getDBTable()->freeBlobHandle(this, blob_id, repo_id, offset, auth_code);
	}
	exit_();
}

bool MSOpenTable::deleteReferences(csWord4 temp_log_id, csWord4 temp_log_offset, bool *must_quit)
{
	MSTableHeadRec		tab_head;
	off_t				blob_id;
	MSTableBlobRec		tab_blob;
	csWord4				repo_id;
	csWord8				repo_offset;
	csWord2				head_size;
	csWord4				auth_code;
	MSRepoFile			*repo_file = NULL;
	bool				result = true;

	enter_();
	openForReading();
	if (myTableFile->read(&tab_head, 0, offsetof(MSTableHeadRec, th_reserved_4), 0) < offsetof(MSTableHeadRec, th_reserved_4))
		/* Nothing to read, delete it ... */
		goto exit;
	if (CS_GET_DISK_4(tab_head.th_temp_log_id_4) != temp_log_id ||
		CS_GET_DISK_4(tab_head.th_temp_log_offset_4) != temp_log_offset) {
		/* Wrong delete reference (ignore): */
		result = false;
		goto exit;
	}

	try_(a) {
		blob_id = CS_GET_DISK_2(tab_head.th_head_size_2);
		while (blob_id + sizeof(MSTableBlobRec) <= getDBTable()->getTableFileSize()) {
			if (*must_quit) {
				/* Bit of a waste of work, but we must quit! */
				result = false;
				break;
			}
			if (myTableFile->read(&tab_blob, blob_id, sizeof(MSTableBlobRec), 0) < sizeof(MSTableBlobRec))
				break;
			repo_id = CS_GET_DISK_3(tab_blob.tb_repo_id_3);
			repo_offset = CS_GET_DISK_6(tab_blob.tb_offset_6);
			head_size = CS_GET_DISK_2(tab_blob.tb_header_size_2);
			auth_code = CS_GET_DISK_4(tab_blob.tb_auth_code_4);
			if (repo_file && repo_file->myRepo->myRepoID != repo_id) {
				repo_file->returnToPool();
				repo_file = NULL;
			}
			if (!repo_file)
				repo_file = getDB()->getRepoFileFromPool(repo_id, true);
			if (repo_file)
				repo_file->freeTableReference(this, repo_offset, head_size, getDBTable()->myTableID, blob_id, auth_code);
			blob_id += sizeof(MSTableBlobRec);
		}
	}
	finally_(a) {
		if (repo_file)
			repo_file->returnToPool();
	}
	cont_(a);

	exit:
	return_(result);
}

void MSOpenTable::openForReading()
{
	if (!myTableFile && !isNotATable)
		myTableFile = getDBTable()->openTableFile();
}

void MSOpenTable::openForWriting()
{
	if (myTableFile && myWriteRepo && myWriteRepoFile)
		return;
	enter_();
	openForReading();
	if (!myWriteRepo || !myWriteRepoFile)
		getDB()->openWriteRepo(this);
	exit_();
}

void MSOpenTable::closeForWriting()
{
	if (myWriteRepoFile) {		
		myWriteRepoFile->myRepo->syncHead(myWriteRepoFile);
		myWriteRepoFile->release();
		myWriteRepoFile = NULL;
	}
	if (myWriteRepo) {
		myWriteRepo->unlockRepo(REPO_WRITE);
#ifndef MS_COMPACTOR_POLLS
		if (myWriteRepo->getGarbageLevel() >= MSRepository::gGarbageThreshold) {
			if (myWriteRepo->myRepoDatabase->myCompactorThread)
				myWriteRepo->myRepoDatabase->myCompactorThread->wakeup();
		}
#endif
		myWriteRepo->release();
		myWriteRepo = NULL;
	}
}

csWord4 MSOpenTable::getTableID()
{
	return myPool->myPoolTable->myTableID;
}

MSTable *MSOpenTable::getDBTable()
{
	return myPool->myPoolTable;
}

MSDatabase *MSOpenTable::getDB()
{
	return myPool->myPoolDB;
}

void MSOpenTable::formatBlobURL(char *blob_url, csWord8 blob_id, csWord4 auth_code, csWord8 blob_size, csWord4 tab_id, csWord8 blob_ref_id)
{
	MSBlobURLRec blob;
	
	blob.bu_type = MS_URL_TYPE_BLOB;
	blob.bu_db_id = getDB()->myDatabaseID;
	blob.bu_tab_id = tab_id;
	blob.bu_blob_id = blob_id;
	blob.bu_auth_code = auth_code;
	blob.bu_server_id = ms_my_get_server_id();
	blob.bu_blob_size = blob_size;
	blob.bu_blob_ref_id = blob_ref_id;
	
	ms_build_blob_url(&blob, blob_url);
	
}
void MSOpenTable::formatBlobURL(char *blob_url, csWord8 blob_id, csWord4 auth_code, csWord8 blob_size, csWord8 blob_ref_id)
{
	formatBlobURL(blob_url, blob_id, auth_code, blob_size, getDBTable()->myTableID, blob_ref_id);
}
void MSOpenTable::formatRepoURL(char *blob_url, csWord4 log_id, csWord8 log_offset, csWord4 auth_code, csWord8 blob_size)
{
	MSBlobURLRec blob;
	
	blob.bu_type = MS_URL_TYPE_REPO;
	blob.bu_db_id = getDB()->myDatabaseID;
	blob.bu_tab_id = log_id;
	blob.bu_blob_id = log_offset;
	blob.bu_auth_code = auth_code;
	blob.bu_server_id = ms_my_get_server_id();
	blob.bu_blob_size = blob_size;
	blob.bu_blob_ref_id = 0;
	
	ms_build_blob_url(&blob, blob_url);
}

MSOpenTable *MSOpenTable::newOpenTable(MSOpenTablePool *pool)
{
	MSOpenTable *otab;
	
	if (!(otab = new MSOpenTable()))
		CSException::throwOSError(CS_CONTEXT, ENOMEM);
	if ((otab->myPool = pool))
		otab->isNotATable = pool->myPoolTable == NULL;
	else
		otab->isNotATable = false;
		
	if (pool && pool->myPoolDB->myBlobCloud)
		otab->myBlobType = MS_CLOUD_PENDING_STORAGE;
	else
		otab->myBlobType = MS_STANDARD_STORAGE;
		
	return otab;
}

/*
 * ---------------------------------------------------------------
 * OPEN TABLE POOLS
 */

MSOpenTablePool::MSOpenTablePool():
myPoolTableID(0),
isRemovingTP(false),
myPoolTable(NULL),
myPoolDB(NULL),
iTablePool(NULL)
{
}

MSOpenTablePool::~MSOpenTablePool()
{
	isRemovingTP = true;
	removeOpenTablesNotInUse();
	/* With this, I also delete those that are in use!: */
	iPoolTables.clear();
	if (myPoolTable)
		myPoolTable->release();
	if (myPoolDB)
		myPoolDB->release();
}

#ifdef DEBUG
void MSOpenTablePool::check()
{
	MSOpenTable	*otab, *ptab;
	bool		found;

	if ((otab = (MSOpenTable *) iPoolTables.getBack())) {
		do {
			found = false;
			ptab = iTablePool;
			while (ptab) {
				if (ptab == otab) {
					ASSERT(!found);
					found = true;
				}
				ptab = ptab->nextTable;
			}
			if (otab->inUse) {
				ASSERT(!found);
			}
			else {
				ASSERT(found);
			}
			otab = (MSOpenTable *) otab->getNextLink();
		} while (otab);
	}
	else
		ASSERT(!iTablePool);
}
#endif

/*
 * This returns the table referenced. So it is safe from the pool being
 * destroyed.
 */
MSOpenTable *MSOpenTablePool::getPoolTable()
{
	MSOpenTable *otab;

	if ((otab = iTablePool)) {
		iTablePool = otab->nextTable;
		otab->nextTable = NULL;
		ASSERT(!otab->inUse);
		otab->inUse = true;
		otab->retain();
	}
	return otab;
}

void MSOpenTablePool::returnOpenTable(MSOpenTable *otab)
{
	otab->inUse = false;
	otab->nextTable = iTablePool;
	iTablePool = otab;
}

/*
 * Add a table to the pool, but do not release it!
 */
void MSOpenTablePool::addOpenTable(MSOpenTable *otab)
{
	iPoolTables.addFront(otab);
}

void MSOpenTablePool::removeOpenTable(MSOpenTable *otab)
{
	otab->close();
	iPoolTables.remove(otab);
}

void MSOpenTablePool::removeOpenTablesNotInUse()
{
	MSOpenTable *otab, *curr_otab;

	iTablePool = NULL;
	/* Remove all tables that are not in use: */
	if ((otab = (MSOpenTable *) iPoolTables.getBack())) {
		do {
			curr_otab = otab;
			otab = (MSOpenTable *) otab->getNextLink();
			if (!curr_otab->inUse)
				iPoolTables.remove(curr_otab);
		} while (otab);
	}
}

void MSOpenTablePool::returnToPool()
{
	MSTableList::removeTablePool(this);
}

MSOpenTablePool *MSOpenTablePool::newPool(csWord4 db_id, csWord4 tab_id)
{
	MSOpenTablePool *pool;
	enter_();
	
	if (!(pool = new MSOpenTablePool())) {
		CSException::throwOSError(CS_CONTEXT, ENOMEM);
	}
	push_(pool);
	pool->myPoolDB = MSDatabase::getDatabase(db_id);
	pool->myPoolTableID = tab_id;
	if (tab_id)
		pool->myPoolTable = pool->myPoolDB->getTable(tab_id, false);
	pop_(pool);
	return_(pool);
}

/*
 * ---------------------------------------------------------------
 * TABLE LIST
 */

CSSyncOrderedList		*MSTableList::gPoolListByID;

MSTableList::MSTableList()
{	
}

MSTableList::~MSTableList()
{
}

void MSTableList::startUp()
{
	new_(gPoolListByID, CSSyncOrderedList);
}

void MSTableList::shutDown()
{
	if (gPoolListByID) {
		gPoolListByID->clear();
		gPoolListByID->release();
		gPoolListByID = NULL;
	}
}

class MSTableKey : public CSOrderKey {
public:
	csWord4	myKeyDatabaseID;
	csWord4	myKeyTableID;

	MSTableKey(): myKeyDatabaseID(0), myKeyTableID(0){ }

	virtual ~MSTableKey() {
	}

	virtual int compareKey(CSOrderKey *x) {
		MSTableKey	*key = (MSTableKey *) x;
		int			r = 0;

		if (myKeyDatabaseID < key->myKeyDatabaseID)
			r = -1;
		else if (myKeyDatabaseID > key->myKeyDatabaseID)
			r = 1;
			
		if (r == 0) {
			if (myKeyTableID < key->myKeyTableID)
				r = -1;
			else if (myKeyTableID > key->myKeyTableID)
				r = 1;
		}
		return r;
	}

public:
	static MSTableKey *newTableKey(csWord4 db_id, csWord4 tab_id)
	{
		MSTableKey *key;

		if (!(key = new MSTableKey())) {
			CSException::throwOSError(CS_CONTEXT, ENOMEM);
		}
		key->myKeyDatabaseID = db_id;
		key->myKeyTableID = tab_id;
		return key;
	}
};

MSOpenTable *MSTableList::getOpenTableByID(csWord4 db_id, csWord4 tab_id)
{
	MSOpenTablePool		*pool;
	MSOpenTable			*otab = NULL;
	MSTableKey			key;

	enter_();
	lock_(gPoolListByID);
	key.myKeyDatabaseID = db_id;
	key.myKeyTableID = tab_id;
	pool = (MSOpenTablePool *) gPoolListByID->find(&key);
	if (!pool) {
		MSTableKey	*key_ptr;
		pool = MSOpenTablePool::newPool(db_id, tab_id);
		key_ptr = MSTableKey::newTableKey(db_id, tab_id);
		gPoolListByID->add(key_ptr, pool);
	}
	if (!(otab = pool->getPoolTable())) {
		otab = MSOpenTable::newOpenTable(pool);
		pool->addOpenTable(otab);
		otab->retain();
	}
	unlock_(gPoolListByID);
	return_(otab);
}

MSOpenTable *MSTableList::getOpenTableForDB(csWord4 db_id)
{
	return(MSTableList::getOpenTableByID(db_id, 0));
}


void MSTableList::releaseTable(MSOpenTable *otab)
{
	MSOpenTablePool	*pool;

	enter_();
	lock_(gPoolListByID);
	push_(otab);
	if ((pool = otab->myPool)) {
		if (pool->isRemovingTP) {
			pool->removeOpenTable(otab);
			gPoolListByID->wakeup();
		}
		else
			pool->returnOpenTable(otab);
	}
	release_(otab);
	unlock_(gPoolListByID);
	exit_();
}

bool MSTableList::removeTablePoolIfEmpty(MSOpenTablePool *pool)
{
	enter_();
	if (pool->getSize() == 0) {
		MSTableKey	key;
		
		key.myKeyDatabaseID = pool->myPoolDB->myDatabaseID;
		key.myKeyTableID = pool->myPoolTableID;
		gPoolListByID->remove(&key);
		/* TODO: Remove the table from the database, if it does not exist
		 * on disk.
		 */
		return_(true);
	}
	return_(false);
}

void MSTableList::removeTablePool(MSOpenTablePool *pool)
{
	enter_();
	lock_(gPoolListByID);
	for (;;) {
		pool->isRemovingTP = true;
		pool->removeOpenTablesNotInUse();
		if (removeTablePoolIfEmpty(pool)) 
			break;

		/*
		 * Wait for the tables that are in use to be
		 * freed.
		 */
		gPoolListByID->wait();
	}
	unlock_(gPoolListByID);
	exit_();
}

/*
 * Close the pool associated with this open table.
 */
void MSTableList::removeTablePool(MSOpenTable *otab)
{
	MSOpenTablePool *pool;
	MSTableKey	key;
	
	key.myKeyDatabaseID = otab->getDB()->myDatabaseID;
	key.myKeyTableID = otab->getTableID();

	enter_();
	frompool_(otab);
	lock_(gPoolListByID);
	for (;;) {
		if (!(pool = (MSOpenTablePool *) gPoolListByID->find(&key)))
			break;
		pool->isRemovingTP = true;
		pool->removeOpenTablesNotInUse();
		if (removeTablePoolIfEmpty(pool))
			break;
		/*
		 * Wait for the tables that are in use to be
		 * freed.
		 */
		gPoolListByID->wait();
	}
	unlock_(gPoolListByID);
	backtopool_(otab);
	exit_();
}

void MSTableList::removeDatabaseTables(MSDatabase *database)
{
	MSOpenTablePool *pool;
	u_int			idx;
	

	enter_();
	push_(database);
	
	retry:
	lock_(gPoolListByID);
	idx = 0;
	while ((pool = (MSOpenTablePool *) gPoolListByID->itemAt(idx))) {
		if (pool->myPoolDB == database) {
			break;
		}
		idx++;
	}
	unlock_(gPoolListByID);

	if (pool) {
		removeTablePool(pool);
		goto retry;
	}
	
	release_(database);
	exit_();
}

// lockTablePoolForDeletion() is only called to lock a pool for a table which is about  to be removed.
// When the pool is returned then it will be removed from the global pool list.
MSOpenTablePool *MSTableList::lockTablePoolForDeletion(csWord4 db_id, csWord4 tab_id, CSString *db_name, CSString *tab_name)
{
	MSOpenTablePool *pool;
	MSTableKey		key;

	enter_();

	push_(db_name);
	if (tab_name)
		push_(tab_name);
		
	key.myKeyDatabaseID = db_id;
	key.myKeyTableID = tab_id;
	
	lock_(gPoolListByID);

	for (;;) {
		if (!(pool = (MSOpenTablePool *) gPoolListByID->find(&key))) {
			char buffer[CS_EXC_MESSAGE_SIZE];

			cs_strcpy(CS_EXC_MESSAGE_SIZE, buffer, "Table is temporarily not available: ");
			cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, db_name->getCString());
			if(tab_name) {
				cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, ".");
				cs_strcat(CS_EXC_MESSAGE_SIZE, buffer, tab_name->getCString());
			}
			CSException::throwException(CS_CONTEXT, MS_ERR_TABLE_LOCKED, buffer);
		}
		pool->isRemovingTP = true;
		pool->removeOpenTablesNotInUse();
		if (pool->getSize() == 0) {
			// pool->retain();	Do not do this. The return to pool will free this by removing it from the list.	
			break;
		}
		/*
		 * Wait for the tables that are in use to be
		 * freed.
		 */
		gPoolListByID->wait();
	}
	unlock_(gPoolListByID);
	
	if (tab_name)
		release_(tab_name);
	release_(db_name);
	return_(pool);	
	
}

MSOpenTablePool *MSTableList::lockTablePoolForDeletion(MSTable *tab)
{
	CSString *tab_name = NULL, *db_name;
	csWord4 db_id, tab_id;
	
	enter_();

	db_name = tab->myDatabase->myDatabaseName;
	db_name->retain();

	tab_name = tab->myTableName;
	tab_name->retain();
	
	db_id = tab->myDatabase->myDatabaseID;
	tab_id = tab->myTableID;
	
	tab->release();
	
	return_( lockTablePoolForDeletion(db_id, tab_id, db_name, tab_name));
}

MSOpenTablePool *MSTableList::lockTablePoolForDeletion(MSOpenTable *otab)
{
	CSString *tab_name = NULL, *db_name;
	csWord4 db_id, tab_id;
	MSTable *tab;

	enter_();
	
	tab = otab->getDBTable();
	if (tab) {
		tab_name = tab->myTableName;
		tab_name->retain();
	}
	
	db_name = otab->getDB()->myDatabaseName;
	db_name->retain();

	db_id = otab->getDB()->myDatabaseID;
	tab_id = otab->getTableID();

	otab->returnToPool();

	return_( lockTablePoolForDeletion(db_id, tab_id, db_name, tab_name));
	
}


