/* Copyright (c) 2009 PrimeBase Technologies GmbH, Germany
 *
 * PrimeBase Media Stream for MySQL
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 * Barry Leslie
 *
 * 2009-05-29
 *
 * H&G2JCtL
 *
 * Repository backup.
 *
 * The backup is done by creating a new database with the same name and ID in the 
 * backup location. Then the pbms_dump table in the source database is initialized
 * for a sequential scan for backup. This has the effect of locking all current repository
 * files. Then the equvalent of  'insert into dst_db.pbms_dump (select * from src_db.pbms_dump);'
 * is performed. 
 *
 */

#include "CSConfig.h"
#include "CSGlobal.h"
#include "CSStrUtil.h"
#include "CSStorage.h"

#include "Defs_ms.h"
#include "SystemTable_ms.h"
#include "OpenTable_ms.h"
#include "Table_ms.h"
#include "Database_ms.h"
#include "Repository_ms.h"
#include "backup_ms.h"
#include "backup_ms.h"
#include "Transaction_ms.h"

MSBackup::MSBackup():
CSDaemon(NULL),
bu_Database(NULL),
bu_BackupList(NULL),
bu_Compactor(NULL),
bu_completed(0),
bu_BackupRunning(false),
bu_size(0)
{
}

MSBackup::~MSBackup()
{
	completeBackup();
		
}

MSBackup *MSBackup::newMSBackup(CSString *backup_location, CSString *db_name, csWord4 db_id)
{
	MSBackup *bu;
	enter_();
	
	new_(bu, MSBackup());
	push_(bu);
	bu->bu_Database = MSDatabase::getBackupDatabase(backup_location, db_name, db_id, true);
	pop_(bu);
	return_(bu);
}

void MSBackup::startBackup(MSDatabase *src_db)
{
	CSSyncVector	*repo_list = src_db->getRepositoryList();
	bool			compacting = false;
	MSRepository	*repo;
	enter_();

	push_(src_db);
	// Build he list of repositories to be backed up.
	lock_(repo_list);

	bu_Compactor = src_db->getCompactorThread();
	if (bu_Compactor) {
		bu_Compactor->retain();
		bu_Compactor->suspend();
	}

	try_(a) {
		new_(bu_BackupList, CSVector(repo_list->size()));
		for (u_int i = 0; i<repo_list->size(); i++) {
			if ((repo = (MSRepository *) repo_list->get(i))) {
				if (!repo->isRemovingFP && !repo->mustBeDeleted) {
					repo->retain();
					bu_BackupList->add(repo);
					if (repo->initBackup() == REPO_COMPACTING) 
						compacting = true; 
					
					if (!repo->myRepoHeadSize) {
						/* The file has not yet been opened, so the
						 * garbage count will not be known!
						 */
						MSRepoFile *repo_file;

						//repo->retain();
						//unlock_(myRepostoryList);
						//push_(repo);
						repo_file = repo->openRepoFile();
						repo_file->release();
						//release_(repo);
						//lock_(myRepostoryList);
						//goto retry;
					}
					
					bu_size += repo->myRepoFileSize; 

				}
			}
		}
		
		// Copy the table list to the backup database:
		csWord4		next_tab = 0;
		MSTable		*tab;
		while ((tab = src_db->getNextTable(&next_tab))) {
			push_(tab);
			bu_Database->addTable(tab->myTableID, tab->myTableName->getCString(), 0, false);
			release_(tab);
		}
		unlock_(repo_list);
		
		// Copy over any physical PBMS system tables.
		bu_Database->retain();
		pop_(src_db);
		pbms_transfer_tables(bu_Database, src_db);
		
		
		// Once the repositories are locked the compactor can be restarted
		// unless it is in the process of compacting a repository that is
		// being backed up.
		if (bu_Compactor && !compacting) {	
			bu_Compactor->resume();		
			bu_Compactor->release();		
			bu_Compactor = NULL;		
		}
		
		// Suspend the transaction writer while the backup is running.
		MSTransactionManager::suspend(true);
		bu_TransactionManagerSuspended = true;
		
		// Start the backup daemon thread.
		bu_ID = bu_start_time = time(NULL);
		start();
	}
	catch_(a) {
		completeBackup();
		throw_();
	}
	cont_(a);
	
	exit_();
}

void MSBackup::completeBackup()
{
	if (bu_TransactionManagerSuspended) {	
		MSTransactionManager::resume();
		bu_TransactionManagerSuspended = false;
	}

	if (bu_BackupList) {
		MSRepository *repo;		
		
		while (bu_BackupList->size()) {
			repo = (MSRepository *) bu_BackupList->take(0);
			if (repo) {				
				repo->backupCompleted();
				repo->release();				
			}
		}
		bu_BackupList->release();
		bu_BackupList = NULL;
	}
		
	if (bu_Compactor) {
		bu_Compactor->resume();
		bu_Compactor->release();
		bu_Compactor = NULL;
	}
	
	if (bu_Database) {
		if (bu_State == BU_COMPLETED)
			bu_Database->releaseBackupDatabase();
		else
			MSDatabase::dropDatabase(bu_Database);
			
		bu_Database = NULL;
	}

	bu_BackupRunning = false;
}

bool MSBackup::doWork()
{
	CSMutex				*lock;
	MSRepository		*src_repo, *dst_repo;
	MSRepoFile			*src_file, *dst_file;
	off_t				src_offset, prev_offset;
	csWord2				head_size;
	csWord8				blob_size, blob_data_size;
	CSStringBuffer		*head;
	MSRepoPointersRec	ptr;
	u_int				table_ref_count;
	u_int				blob_ref_count;
	int					ref_count;
	size_t				ref_size;
	csWord4				auth_code;
	csWord4				tab_id;
	csWord8				blob_id;
	MSOpenTable			*otab;
	csWord4				src_repo_id;
	csWord1				status;
	csWord2				tab_index;
	csWord4				mod_time;
	MSRepoTempRefPtr	cloud_ref;
	char				transferBuffer[MS_BACKUP_BUFFER_SIZE];
	
	enter_();
	bu_BackupRunning = true;
	bu_State = BU_RUNNING; 

/*
	// For testing:
	{
		int blockit = 0;
		myWaitTime = 5 * 1000;  // Time in milli-seconds
		while (blockit)
			return_(true);
	}
*/
	
	try_(a) {
		new_(head, CSStringBuffer(100));
		push_(head);

		src_repo = (MSRepository*)bu_BackupList->get(0);
		while (src_repo && !myMustQuit) {
			src_offset = 0;
			src_file = src_repo->openRepoFile();
			push_(src_file);

			dst_repo = bu_Database->lockRepo(src_repo->myRepoFileSize - src_repo->myGarbageCount);
			frompool_(dst_repo);
			dst_file = dst_repo->openRepoFile();
			push_(dst_file);
			
			src_repo_id = src_repo->myRepoID;
			src_offset = src_repo->myRepoHeadSize;
			prev_offset = 0;
			while (src_offset < src_repo->myRepoFileSize) {	
	retry_read:
					
				bu_completed += src_offset - prev_offset;
				prev_offset = src_offset;
				suspended();

				if (myMustQuit)
					break;
				
				// A lock is required here because references and dereferences to the
				// BLOBs can result in the repository record being updated while 
				// it is being copied.
				lock = &src_repo->myRepoLock[src_offset % CS_REPO_REC_LOCK_COUNT];
				lock_(lock);
				head->setLength(src_repo->myRepoBlobHeadSize);
				if (src_file->read(head->getBuffer(0), src_offset, src_repo->myRepoBlobHeadSize, 0) < src_repo->myRepoBlobHeadSize) { 
					unlock_(lock);
					break;
				}
					
				ptr.rp_chars = head->getBuffer(0);
				ref_size = CS_GET_DISK_1(ptr.rp_head->rb_ref_size_1);
				ref_count = CS_GET_DISK_2(ptr.rp_head->rb_ref_count_2);
				head_size = CS_GET_DISK_2(ptr.rp_head->rb_head_size_2);
				blob_size = CS_GET_DISK_6(ptr.rp_head->rb_blob_size_6);
				blob_data_size = CS_GET_DISK_6(ptr.rp_head->rb_blob_data_size_6);
				auth_code = CS_GET_DISK_4(ptr.rp_head->rb_auth_code_4);
				status = CS_GET_DISK_1(ptr.rp_head->rb_status_1);
				mod_time = CS_GET_DISK_4(ptr.rp_head->rb_mod_time_4);
				
				// If the BLOB was modified after the start of the backup
				// then set the mod time to the abckup time to ensure that
				// a backup for update will work correctly.
				if (mod_time > bu_start_time)
					CS_SET_DISK_4(ptr.rp_head->rb_mod_time_4, bu_start_time);
					
				// If the BLOB was moved during the time of this backup then copy
				// it to the backup location as a referenced BLOB.
				if ((status == MS_BLOB_MOVED)  && (bu_ID == (csWord4) CS_GET_DISK_4(ptr.rp_head->rb_backup_id_4))) {
					status = MS_BLOB_REFERENCED;
					CS_SET_DISK_1(ptr.rp_head->rb_status_1, status);
				}
				
				// sanity check
				if ((blob_data_size == 0) || ref_count <= 0 || ref_size == 0 ||
					head_size < src_repo->myRepoBlobHeadSize + ref_count * ref_size ||
					!VALID_BLOB_STATUS(status)) {
					/* Can't be true. Assume this is garbage! */
					src_offset++;
					unlock_(lock);
					continue;
				}
				
				
				if ((status == MS_BLOB_REFERENCED) || (status == MS_BLOB_MOVED)) {
					head->setLength(head_size);
					if (src_file->read(head->getBuffer(0) + src_repo->myRepoBlobHeadSize, src_offset + src_repo->myRepoBlobHeadSize, head_size  - src_repo->myRepoBlobHeadSize, 0) != (head_size- src_repo->myRepoBlobHeadSize)) {
						unlock_(lock);
						break;
					}

					table_ref_count = 0;
					blob_ref_count = 0;
					cloud_ref = NULL;
					
					// Loop through all the references removing temporary references 
					// and counting table and blob references.
					
					ptr.rp_chars = head->getBuffer(0) + src_repo->myRepoBlobHeadSize;
					for (int count = 0; count < ref_count; count++) {
						switch (CS_GET_DISK_2(ptr.rp_ref->rr_type_2)) {
							case MS_BLOB_FREE_REF:
								break;
							case MS_BLOB_TABLE_REF:
								// Unlike the compactor, table refs are not checked because
								// they do not yet exist in the backup database.
								table_ref_count++;
								break;
							case MS_BLOB_DELETE_REF:
								// These are temporary references from the TempLog file. 
								// They are not copied to the backup. 
								CS_SET_DISK_2(ptr.rp_ref->rr_type_2, MS_BLOB_FREE_REF);
								break;
							case MS_BLOB_CLOUD_REF:
								if (!cloud_ref)
									cloud_ref = ptr.rp_temp_ref;
								else { // There shouldn't be more than one of these per BLOB.
									CS_SET_DISK_2(ptr.rp_ref->rr_type_2, MS_BLOB_FREE_REF);
								}

								break;
							default:
								// Must be a BLOB reference
								
								tab_index = CS_GET_DISK_2(ptr.rp_blob_ref->er_table_2);
								if (tab_index && (tab_index <= ref_count)) {
									// Only committed references are backed up.
									if (IS_COMMITTED(CS_GET_DISK_8(ptr.rp_blob_ref->er_blob_ref_id_8))) {
										MSRepoTableRefPtr	tab_ref;
										tab_ref = (MSRepoTableRefPtr) (head->getBuffer(0) + src_repo->myRepoBlobHeadSize + (tab_index-1) * ref_size);
										if (CS_GET_DISK_2(tab_ref->rr_type_2) == MS_BLOB_TABLE_REF)
											blob_ref_count++;
									} else {
										CS_SET_DISK_2(ptr.rp_ref->rr_type_2, MS_BLOB_FREE_REF);
									}
								
								} else {
									/* Can't be true. Assume this is garbage! */
									src_offset++;
									unlock_(lock);
									goto retry_read;
								}
								break;
						}
						ptr.rp_chars += ref_size;
					}


					// If there are still blob references then the record needs to be backed up.
					if (table_ref_count && blob_ref_count) {

						off_t dst_offset;

						dst_offset = dst_repo->myRepoFileSize;
						
						// If there is a cloud event it must be created before the header
						// is written because it will update the header.
						if (cloud_ref) {
							csWord4		log_id; 
							csWord4		log_offset;
							csWord4		temp_time;

							otab = MSTableList::getOpenTableForDB(bu_Database->myDatabaseID);
							frompool_(otab);
							bu_Database->queueCloudEvent(otab, dst_repo->myRepoID, dst_offset, auth_code, &log_id, &log_offset, &temp_time);
							backtopool_(otab);
							dst_repo->myLastTempTime = temp_time;
							CS_SET_DISK_4(cloud_ref->tp_log_id_4, log_id);
							CS_SET_DISK_4(cloud_ref->tp_offset_4, log_offset);
						}

						/* Write the header. */
						dst_file->write(head->getBuffer(0), dst_offset, head_size);

						/* Copy the BLOB over: */
						CSFile::transfer(dst_file, dst_offset + head_size, src_file, src_offset + head_size, blob_size, transferBuffer, MS_BACKUP_BUFFER_SIZE);
					
						/* Update the references: */
						ptr.rp_chars = head->getBuffer(0) + src_repo->myRepoBlobHeadSize;
						for (int count = 0; count < ref_count; count++) {
							switch (CS_GET_DISK_2(ptr.rp_ref->rr_type_2)) {
								case MS_BLOB_FREE_REF:
								case MS_BLOB_DELETE_REF:
								case MS_BLOB_CLOUD_REF:
									break;
								case MS_BLOB_TABLE_REF:
									tab_id = CS_GET_DISK_4(ptr.rp_tab_ref->tr_table_id_4);
									blob_id = CS_GET_DISK_6(ptr.rp_tab_ref->tr_blob_id_6);

									if ((otab = MSTableList::getOpenTableByID(bu_Database->myDatabaseID, tab_id))) {
										frompool_(otab);
										otab->getDBTable()->setBlobHandle(otab, blob_id, dst_repo->myRepoID, dst_offset, blob_size, head_size, auth_code);
//CSException::throwException(CS_CONTEXT, MS_ERR_NOT_IMPLEMENTED, "What if an error ocurred here!");

										backtopool_(otab);
									}
									break;
								default:
									break;
							}
							ptr.rp_chars += ref_size;
						}

						dst_repo->myRepoFileSize += head_size + blob_size;
					}
				}
				unlock_(lock);
				src_offset += head_size + blob_size;
			}
			bu_completed += src_offset - prev_offset;
			
			// close the destination repository and cleanup.
			release_(dst_file);
			backtopool_(dst_repo);
			release_(src_file);
			
			// release the source repository and get the next one in the list.
			src_repo->backupCompleted();
			bu_BackupList->remove(0);
			
			src_repo = (MSRepository*)bu_BackupList->get(0);
		}
				
		release_(head);
		if (myMustQuit)
			bu_State = BU_TERMINATED; 
		else
			bu_State = BU_COMPLETED; 
			
	}	
	
	catch_(a) {
		logException();
	}
	
	cont_(a);	
	completeBackup();
	myMustQuit = true;
	return_(true);
}

void *MSBackup::finalize()
{
	completeBackup();
	return NULL;
};
